<?php
/**
 * @file
 * Administration functions for i18n_auto_clone.
 */

/**
 * Administration form for setting defaults.
 */
function i18n_auto_clone_defaults_settings_form(&$form_state) {
  $form['i18n_auto_clone_node_types'] = array(
    '#type' => 'checkboxes',
    '#multiple' => TRUE,
    '#title' => t('Content types to keep dummy translations for'),
    '#options' => array(),
    '#default_value' => variable_get('i18n_auto_clone_node_types', array()),
    '#description' => t('The content types selected will have translations generated by a cloning procedure. These nodes will be marked as needing a translation update. When changes are made to the default translation in the set, the dummy clones will be updated.'),
    '#size' => 15,
  );
  foreach (node_get_types('names') as $type => $name) {
    $form['i18n_auto_clone_node_types']['#options'][$type] = check_plain($name);
  }
  $languages = locale_language_list('name', TRUE);
  // Unset the language set as default.
  unset($languages[variable_get('i18n_auto_clone_default_language', 'en')]);

  $form['i18n_auto_clone_enabled_languages'] = array(
    '#type' => 'checkboxes',
    '#multiple' => TRUE,
    '#title' => t('Languages to enable for dummy translations'),
    '#options' => $languages,
    '#default_value' => variable_get('i18n_auto_clone_enabled_languages', array()),
    '#description' => t('The content types selected will have translations generated by a cloning procedure. These nodes will be marked as needing a translation update. When changes are made to the default translation in the set, the dummy clones will be updated.'),
    '#size' => 15,
  );
  $form['i18n_auto_clone_default_language'] = array(
    '#type' => 'select',
    '#multiple' => FALSE,
    '#title' => t('Default language'),
    '#default_value' => variable_get('i18n_auto_clone_default_language', 'en'),
    '#options' => locale_language_list('name', TRUE),
    '#description' => t('Choose the default language that will act as the master node for the translation clone set, sitewide. Creating a new node (of an enabled type) in this language will set it as the "master node" and new clones will be created for all enabled languages.'),
  );
  $form['i18n_auto_clone_processing'] = array(
    '#type' => 'radios',
    '#title' => t('Clone processing'),
    '#default_value' => variable_get('i18n_auto_clone_processing', 'during_cron'),
    '#options' => array(
      'during_cron' => t('During cron'),
      'immediately' => t('Immediately upon node save'),
    ),
  );
  $form['i18n_auto_clone_cron_limit'] = array(
    '#type' => 'textfield',
    '#title' => t('Cron processing limit'),
    '#default_value' => variable_get('i18n_auto_clone_cron_limit', 10),
    '#size' => 20,
    '#required' => TRUE,
  );
  $form['i18n_auto_clone_batch_size'] = array(
    '#type' => 'textfield',
    '#title' => t('Cloning batch size'),
    '#default_value' => variable_get('i18n_auto_clone_batch_size', 10),
    '#size' => 20,
    '#required' => TRUE,
    '#description' => t('Number of clones to queue and/or create per batch cycle. Increasing this value may speed up the batch process at the expense of server resources, and may result in fatal errors if the PHP memory limit is not high enough. The queueing process is much less intensive, so this number can be very high for that purpose.'),
  );
  $form['buttons']['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Save configuration'),
  );
  $form['#theme'] = 'system_settings_form';
  return system_settings_form($form);
}


/**
 * Form callback. Batch process utility.
 */
function i18n_auto_clone_batch_form() {
  $immediate_processing = variable_get('i18n_auto_clone_processing', 'during_cron') == 'immediately' ? TRUE : FALSE;
  $types = array();
  $enabled_types = variable_get('i18n_auto_clone_node_types', array());
  foreach (node_get_types('names') as $type => $name) {
    if (isset($enabled_types[$type]) && $enabled_types[$type]) {
      $types[$type] = check_plain($name);
    }
  }
  $form['i18n_auto_clone_batch_node_types'] = array(
    '#type' => 'checkboxes',
    '#title' => t('Content types to batch process'),
    '#options' => $types,
    '#default_value' => array(),
    '#required' => TRUE,
    '#description' => t('The content types selected will have translations generated by a cloning procedure. These nodes will be marked as needing a translation update. When changes are made to the default translation in the set, the dummy clones will be updated. Note that the node types will not be available here unless they are enabled on the settings overview page.'),
  );
  $batch_languages = array();
  $languages = locale_language_list('name', TRUE);
  $enabled_languages = variable_get('i18n_auto_clone_enabled_languages', array());
  foreach ($languages as $type => $name) {
    if ($enabled_languages[$type]) {
      $batch_languages[$type] = check_plain($name);
    }
  }
  $form['i18n_auto_clone_batch_languages'] = array(
    '#type' => 'checkboxes',
    '#title' => t('Languages to batch process'),
    '#options' => $batch_languages,
    '#default_value' => array(),
    '#required' => TRUE,
    '#description' => t('The languages selected will have translations generated by a cloning procedure. These nodes will be marked as needing a translation update. When changes are made to the default translation in the set, the dummy clones will be updated. Note that the languages will not be available here unless they are enabled on the settings overview page.'),
  );
  $form['queue'] = array(
    '#type' => 'submit',
    '#value' => t('Queue nodes for cloning'),
    '#suffix' => '<p>' . t('This will queue all the missing tranlations for the selected node types.') . '</p>',
  );
  $form['clone'] = array(
    '#type' => 'submit',
    '#value' => t('Create or update clones'),
    '#suffix' => '<p>' . t('This will create or update all clones currently queued for cloning.') . '</p>',
  );
  $immediate_title = t('Queue nodes for cloning and clone them');

  if ($immediate_processing) {
    $form['queue']['#value'] = t('Queue nodes for cloning and clone them');
    $form['queue']['#suffix'] = '<p>' . t('This will queue all the missing tranlations for the selected node types, and execute the cloning. The two processes are combined because immediate processing is selected on the Overview configuration page.') . '</p>';
  }

  return $form;
}

// @TODO: Form validation. (they need to select at least one of each option.
/**
 * Batch form submit handler function.
 *
 * Sets up batch processeses based on the submit button pressed.
 */
function i18n_auto_clone_batch_form_submit($form, &$form_state) {
  $languages = array();
  foreach ($form_state['values']['i18n_auto_clone_batch_languages'] as $language => $enabled) {
    if ($enabled) {
      $languages[] = $language;
    }
  }
  $node_types = array();
  foreach ($form_state['values']['i18n_auto_clone_batch_node_types'] as $node_type => $enabled) {
    if ($enabled) {
      $node_types[] = $node_type;
    }
  }
  $options = array(
    'node_types' => $node_types,
    'languages' => $languages,
    'operation' => $form_state['values']['op'],
  );
  if ($form_state['values']['op'] == t('Queue nodes for cloning') || $form_state['values']['op'] == t('Queue nodes for cloning and clone them')) {
    $batch = array(
      'title' => $form_state['values']['op'],
      'operations' => array(
        array(
          'i18n_auto_clone_batch_queue',
          array($options),
        ),
      ),
      'finished' => 'i18n_auto_clone_batch_finished',
      'file' => drupal_get_path('module', 'i18n_auto_clone') . '/i18n_auto_clone.admin.inc',
    );
  }
  else {
    $batch = array(
      'title' => $form_state['values']['op'],
      'operations' => array(
        array(
          'i18n_auto_clone_batch_clone',
          array($options),
        ),
      ),
      'finished' => 'i18n_auto_clone_batch_finished',
      'file' => drupal_get_path('module', 'i18n_auto_clone') . '/i18n_auto_clone.admin.inc',
    );
  }
  batch_set($batch);
}

/**
 * Batch function to add nodes to the queue.
 */
function i18n_auto_clone_batch_queue($options, &$context) {
  // We'll need to load this include file.
  module_load_include('inc', 'node', 'node.pages');

  if (!isset($context['sandbox']['progress'])) {
    $context['sandbox']['progress'] = 0;
    $context['sandbox']['current_result'] = 0;
  }
  else {
    // Keep track of previous cycle's progress so we don't get stuck in a loop.
    $context['sandbox']['previous_progress'] = $context['sandbox']['progress'];
  }
  // We need a query to count number of rows.
  if (!isset($context['sandbox']['max'])) {
    $context['sandbox']['max'] = _i18n_auto_clone_batch_queue_number_left($options['node_types'], $options['languages']);
  }

  // Loop through x number of rows at a time.
  $batch_size = variable_get('i18n_auto_clone_batch_size', 5);
  $default_language = variable_get('i18n_auto_clone_default_language', 'en');

  // Node must match node-type and default language.
  // Specifications for the subquery, which excludes certain nodes, based on
  // tnid:
  // -- language/node-type matches AND clone record does exist.
  // OR
  // -- language/node-type matches AND Translation exists AND clone record does not.
  $result = db_query_range("SELECT DISTINCT n.nid FROM {node} n
    LEFT JOIN {i18n_auto_clone} clone ON n.nid = clone.tnid
    WHERE n.type IN (" . db_placeholders($options['node_types'], 'varchar') . ")
    AND n.language = '%s'
    AND n.nid NOT IN(
      SELECT DISTINCT n.nid FROM {node} n
      LEFT JOIN {i18n_auto_clone} clone ON n.nid = clone.tnid
      WHERE (n.type IN(" . db_placeholders($options['node_types'], 'varchar') . ")
      AND clone.language IN(" . db_placeholders($options['languages'], 'varchar') . ")
      AND clone.tnid = n.nid)
      OR (n.type IN(" . db_placeholders($options['node_types'], 'varchar') . ")
      AND n.language IN(" . db_placeholders($options['languages'], 'varchar') . ")
      AND clone.tnid IS NULL)
    )
    ORDER BY n.nid", array_merge($options['node_types'], (array) $default_language, $options['node_types'], $options['languages'], $options['node_types'], $options['languages']), 0, $batch_size);
  // Keep track of how many nids the query returns.
  $count = 0;
  $nids = array();
  while ($row = db_fetch_object($result)) {
    $nids[] = $row->nid;
    $count ++;
  }
  $op = $options['operation'] == t('Queue nodes for cloning') ? 'queued' : 'queued and cloned';
  // TODO: Catch an exception where there were 0 results all along.
  foreach ($nids as $nid) {
    $node = node_load($nid);
    foreach ($options['languages'] as $language) {
      $result = i18n_auto_clone_queue_clone($node, $language);
      if ($result && $result != 'duplicate') {
        if ($op == 'queued') {
          $context['results'][] = t('%language clone for nid:%nid successfully queued.', array('%language' => $language, '%nid' => $nid));
        }
        else {
          $context['results'][] = t('%language clone for nid:%nid successfully queued and cloned. Clone nid: %clone', array(
              '%language' => $language,
              '%nid' => $nid,
              '%clone' => $result,
            ));
        }
      }
    }
  }

  // Update our progress information.
  $context['sandbox']['progress'] += $count;
  // Add a faux-count to the progress if we're stuck in a loop.
  if ($context['sandbox']['previous_progress'] === $context['sandbox']['progress']) {
    $context['sandbox']['progress']++;
  }
  $context['message'] = t('Now processing %current of %max nodes to be potentially added to the cloning queue.', array(
    '%current'=> $context['sandbox']['progress'],
    '%max' => $context['sandbox']['max'],
  ));

  // Inform the batch engine that we are not finished,
  // and provide an estimation of the completion level we reached.
  if ($context['sandbox']['progress'] != $context['sandbox']['max']) {
    $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
  }

}

/**
 * Query to find number of nodes left to add to the queue.
 */
function _i18n_auto_clone_batch_queue_number_left($node_types, $languages) {
  $default_language = variable_get('i18n_auto_clone_default_language', 'en');
  $number_to_process = db_result(db_query("SELECT DISTINCT COUNT(n.nid) FROM {node} n
    LEFT JOIN {i18n_auto_clone} clone ON n.nid = clone.tnid
    WHERE n.type IN (" . db_placeholders($node_types, 'varchar') . ")
    AND n.language = '%s'
    AND n.nid NOT IN(
      SELECT DISTINCT n.nid FROM {node} n
      LEFT JOIN {i18n_auto_clone} clone ON n.nid = clone.tnid
      WHERE (n.type IN(" . db_placeholders($node_types, 'varchar') . ")
      AND clone.language IN(" . db_placeholders($languages, 'varchar') . ")
      AND clone.tnid = n.nid)
      OR (n.type IN(" . db_placeholders($node_types, 'varchar') . ")
      AND n.language IN(" . db_placeholders($languages, 'varchar') . ")
      AND clone.tnid IS NULL)
    )", array_merge($node_types, (array) $default_language, $node_types, $languages, $node_types, $languages)));
  return $number_to_process;
}
/**
 * Batch function to clone nodes already in the queue.
 */
function i18n_auto_clone_batch_clone($options, &$context) {
  // We'll need to load this include file.
  module_load_include('inc', 'node', 'node.pages');

  if (!isset($context['sandbox']['progress'])) {
    $context['sandbox']['progress'] = 0;
    $context['sandbox']['current_result'] = 0;
  }
  else {
    // Keep track of previous cycle's progress so we don't get stuck in a loop.
    $context['sandbox']['previous_progress'] = $context['sandbox']['progress'];
  }

  // We query to count number of rows and set it on the first cycle.
  if (!isset($context['sandbox']['max'])) {
    $context['sandbox']['max'] = _i18n_auto_clone_batch_clone_number_left($options['node_types'], $options['languages']);
  }

  // Loop through a variable number of rows per batch cycle.
  $batch_size = variable_get('i18n_auto_clone_batch_size', 5);
  $default_language = variable_get('i18n_auto_clone_default_language', 'en');
  $result = db_query_range("SELECT clone.nid, clone.tnid, clone.language FROM {i18n_auto_clone} clone
    LEFT JOIN {node} n ON clone.tnid = n.nid
    WHERE clone.needs_update = 1
    AND n.type IN (" . db_placeholders($options['node_types'], 'varchar') . ")
    AND clone.language IN (" . db_placeholders($options['languages'], 'varchar') . ")", array_merge($options['node_types'], $options['languages']), 0, $batch_size);
  $count = 0;
  while ($clone = db_fetch_object($result)) {
    $original_node = node_load($clone->tnid);
    $clones[] = array(
      'original_node' => $original_node,
      'language' => $clone->language,
      'clone_nid' => $clone->nid,
    );
    $count++;
  }
  foreach ($clones as $clone) {
    if ($nid = i18n_auto_clone_create_clone($clone['original_node'], $clone['language'], $clone['clone_nid'])) {
      $context['results'][] = t('%language clone for nid:%nid successfully spawned.', array('%language' => $language, '%nid' => $nid));
    }

  }
  // Update our progress information.
  $context['sandbox']['progress'] += $count;
  // Add a faux-count to the progress if we're stuck in a loop.
  if ($context['sandbox']['previous_progress'] === $context['sandbox']['progress']) {
    $context['sandbox']['progress'] ++;
  }
  $context['message'] = t('Now creating %current of %max clones.', array(
    '%current'=> $context['sandbox']['progress'],
    '%max' => $context['sandbox']['max'],
  ));

  // Inform the batch engine that we are not finished,
  // and provide an estimation of the completion level we reached.
  if ($context['sandbox']['progress'] != $context['sandbox']['max']) {
    $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
  }
}

/**
 * Query to find number of entries in the queue left to clone.
 */
function _i18n_auto_clone_batch_clone_number_left($node_types, $languages) {
  // Join on node's nid column, since clones may not yet exist in the node table.
  $result = db_result(db_query("SELECT COUNT(clone.tnid) FROM {i18n_auto_clone} clone
    LEFT JOIN {node} n ON clone.tnid = n.nid
    WHERE clone.needs_update = 1
    AND n.type IN (" . db_placeholders($node_types, 'varchar') . ")
    AND clone.language IN (" . db_placeholders($languages, 'varchar') . ")", array_merge($node_types, $languages)));
  return $result;
}

/**
 * End of batch function.
 */
function i18n_auto_clone_batch_finished($success, $results, $operations) {
  if ($success) {
    // Here we do something meaningful with the results.
    $message = count($results) . ' items processed.';
    $message .= theme('item_list', $results);
  }
  else {
    // An error occurred.
    // $operations contains the operations that remained unprocessed.
    $error_operation = reset($operations);

    $message = t('An error occurred while processing %error_operation with arguments: @arguments', array('%error_operation' => $error_operation[0], '@arguments' => print_r($error_operation[1], TRUE)));
    watchdog('i18_auto_clone', 'i18_auto_clone encountered an error during batch batch. The operation name is %error_operation and arguments are @arguments.', array(
        '%error_operation' => $error_operation[0],
        '@arguments' => print_r($error_operation[1]),
      ), WATCHDOG_ERROR);
  }

  drupal_set_message($message);
}
